/**
 * 
 * Multiple 18650 Battery Capacity Tester
 * Original Idea is taken from this guy
 * https://www.instructables.com/id/Rechargeable-Battery-Capacity-Tester/
 * and most of the Arduino code is from him
 */

#include "SPI.h"
#include "Adafruit_GFX.h"
#include "Adafruit_ILI9341.h"

/** colors */

#define BLACK    0x0000
#define BLUE     0x001F
#define RED      0xFF00
#define GREEN    0x07E0
#define CYAN     0x07FF
#define MAGENTA  0xF81F
#define YELLOW   0xFFE0 
#define WHITE    0xFFFF
#define GREY     0x2104

// Meter colour schemes
#define RED2RED 0
#define GREEN2GREEN 1
#define BLUE2BLUE 2
#define BLUE2RED 3
#define GREEN2RED 4
#define RED2GREEN 5

#define BG_color BLUE

#include <Fonts/FreeMonoBoldOblique12pt7b.h>
#include <Fonts/FreeSerif9pt7b.h>
#include <Fonts/FreeSans9pt7b.h>
// For the Adafruit shield, these are the default.
#define TFT_DC 9
#define TFT_CS 10

// Use hardware SPI (on Uno, #13, #12, #11) and the above for CS/DC
Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS, TFT_DC);
// If using the breakout, change pins as desired
//Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS, TFT_DC, TFT_MOSI, TFT_CLK, TFT_RST, TFT_MISO);

#define LCD_LINE1            0
#define LCD_LINE2            1
#define LCD_LINE3            2
#define LCD_LINE4            3
#define LCD_LINE5            4
#define LCD_LINE6            5
#define NUM_LINES_ON_LCD     6

#define MAX_BATTERIES        4
#define LOAD_RESISTANCE      0.333    //load resistance in ohms (3x1 Ohm SMD Resistors in Paralel as Shunt)
#define START_BEEP        5000      // frequency in Hz
#define DONE_BEEP         1800
#define CHECK_BEEP        3000
#define SPKR_PIN             7      // Buzzer Pin
#define BTN_PIN              6      // Button Pin
#define BATTERY_ICON_HORIZ  34     // Horizontal position of battery icon (in pixels)
#define MAH_HORIZ_POSITION  60     // Horizontal position of mAh display


#define MAX_VOLTAGE      4290   // Max Voltage used for detection (in mV)
#define MIN_VOLTAGE      2900   //  Min Voltage for load removal (in mV)

#define schema              4   // desired colour schema

static const byte batteryEmptyIcon[] ={ 0x1c, 0x14, 0x77,0x41,0x41,0x41,
           0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x7f};
static const byte batteryFullIcon[] = { 0x1c, 0x14, 0x77,0x7f,0x7f,0x7f,
           0x7f,0x7f,0x7f,0x7f,0x7f,0x7f,0x7f,0x7f,0x7f,0x7f,0x7f,0x7f};
static const byte noBatteryIcon [] = { 0x1C, 0x14, 0x77, 0x41, 0x41,0x41,
           0x41,0x63,0x77,0x5D,0x5D,0x77,0x63,0x41,0x41,0x41,0x41,0x7f};

unsigned long Time;
byte seconds ;
byte minutes ;
byte hours ;

struct batteryStruct
{
    unsigned long  charge;         // Total microamp hours for this battery
    byte battStatus;               // set this to DONE when this cell's test is complete
    byte batteryVoltagePin;        // Analog sensor pin (0-5) for reading battery voltage
    byte fetVoltagePin;            // Analog sensor pin (0-5) to read voltage across FET
    byte dischargeControlPin;      // Output Pin that controlls the load for this battery
    unsigned int lowerThreshold;   // voltage at which discharge is complete (mV)
    unsigned long PrevTime;        // Previous time reading (in milliseconds)
    unsigned int numSamplesAboveMin; // number of good voltage readings (to determine battery installed)
    unsigned int numSamplesBelowMin; // number of samples read below minimum (to determine battery discharged)   
    int oldVolt;                     // last measured voltage on Battery, value on display refreshes only if measured value has changed
    int oldCap;                      // sam as above, for capacity
    float BatResRatio;              // Voltage Divider ratio (Battery side)
    float FetResRatio;              // Voltage Divider ratio (Resistor side)
    unsigned long RunTime;          // Battery testing time (in milliseconds)
}
battery[MAX_BATTERIES];


void printCharge(unsigned int n, unsigned short numDigits, int bat);
void ClearDisplayLine(int line);   // function to clear specified line of LCD display
unsigned int getBatteryVoltage(unsigned int batteryNum);
unsigned int getFetVoltage(unsigned int batteryNum);
void printVoltage(unsigned int n);

// external reference Voltage -> measured on PIN 3.3V used as refV
//float refV = 2938.;  // in mV (2.938V - Arduino powered by PC USB)
float refV = 2945.;  // in mV (2.945V - Arduino powered by 5V adapter)

enum {NOT_INSTALLED, DETECTING_TYPE, OVER_VOLTAGE, TEST_IN_PROGRESS, DONE};  // battery status values

uint32_t runTime = -99999;       // time for next update

void setup() {
  analogReference(EXTERNAL); // use AREF for reference voltage - connected to 3.3V PIN
  //refV = readVcc();  // good function, but not accurate
  //Serial.begin(9600);
  Serial.print("V-REF: ");
  Serial.println(refV);
  unsigned int batteryNum;
  unsigned int battVoltage;

  pinMode(SPKR_PIN, OUTPUT);      //Set output mode for pin used for spkr
  pinMode(BTN_PIN, INPUT);        // input mode for button pin
  
  battery[0].dischargeControlPin = 2;    // setup the corresponding pins
   battery[0].fetVoltagePin = A7;        // for each battery according to
   battery[0].batteryVoltagePin = A6;    // schematic wiring diagram.
   battery[0].oldVolt=0.0;               
   battery[0].oldCap=0.0;                
   // measure your own resistors ratio by measuring voltage on voltage dividers. 
   // use 1% tollerance resistors
   battery[0].BatResRatio=3.1280;   // Battery voltage divided by Voltage on batteryVoltagePin
   battery[0].FetResRatio=3.1282;   // Battery voltage divided by Voltage on fetVoltagePin 
   
   battery[1].dischargeControlPin = 3;       
   battery[1].fetVoltagePin = A5; 
   battery[1].batteryVoltagePin = A4; 
   battery[1].oldVolt=0.0;
   battery[1].oldCap=0.0;
   battery[1].BatResRatio=3.1317;
   battery[1].FetResRatio=3.1286;
   
   battery[2].dischargeControlPin = 4;
   battery[2].fetVoltagePin = A3; 
   battery[2].batteryVoltagePin = A2; 
   battery[2].oldVolt=0.0;
   battery[2].oldCap=0.0;
   battery[2].BatResRatio=3.1300;
   battery[2].FetResRatio=3.1307;
      
   battery[3].dischargeControlPin = 5;
   battery[3].fetVoltagePin = A1; 
   battery[3].batteryVoltagePin = A0;
   battery[3].oldVolt=0.0;
   battery[3].oldCap=0.0;
   battery[3].BatResRatio=3.1258;
   battery[3].FetResRatio=3.1291;
     
   battery[0].battStatus = DETECTING_TYPE;  // initialize status of each battery
   battery[1].battStatus = DETECTING_TYPE; 
   battery[2].battStatus = DETECTING_TYPE;
   battery[3].battStatus = DETECTING_TYPE;
   
   battery[0].numSamplesAboveMin = 0;
   battery[1].numSamplesAboveMin = 0;
   battery[2].numSamplesAboveMin = 0;
   battery[3].numSamplesAboveMin = 0;
  
   battery[0].numSamplesBelowMin = 0;
   battery[1].numSamplesBelowMin = 0;
   battery[2].numSamplesBelowMin = 0; 
   battery[3].numSamplesBelowMin = 0; 

   
   // Set the three FET control pins for output
   pinMode(battery[0].dischargeControlPin, OUTPUT); 
   pinMode(battery[1].dischargeControlPin, OUTPUT); 
   pinMode(battery[2].dischargeControlPin, OUTPUT); 
   pinMode(battery[3].dischargeControlPin, OUTPUT);
   
   digitalWrite(battery[0].dischargeControlPin, LOW);
   digitalWrite(battery[1].dischargeControlPin, LOW);
   digitalWrite(battery[2].dischargeControlPin, LOW);
   digitalWrite(battery[3].dischargeControlPin, LOW);
  tft.begin();
  tft.setRotation(1);
  splash();
  delay(3000);
  tft.fillScreen(BG_color);
  mainScreen();
}

void loop() 
{
   static unsigned int i = 0 ;
   static unsigned int   p, batteryNum, width1;
   static unsigned long duration, currentTime, runningTime;
   static unsigned int battVoltage,fetVoltage, loadCurrent;
   static boolean done = false;
   static unsigned int beepCounter = 0;
  
   if (!done)
   {
      for (batteryNum= 0 ; batteryNum < MAX_BATTERIES ; batteryNum++)
      {
         battVoltage=fetVoltage=0.0;
         battVoltage = getBatteryVoltage(batteryNum);
         fetVoltage = getFetVoltage(batteryNum);

        if(Serial){
             Serial.print("V-bat:");
             Serial.println(battVoltage);
             Serial.print("V-FET:");
             Serial.println(analogRead( fetVoltage));
             
         }
            
         if ( battery[batteryNum].battStatus == TEST_IN_PROGRESS)
         {
            if(battery[batteryNum].oldVolt != battVoltage){
              ClearDisplayLine(batteryNum);
              printVoltage(battVoltage,batteryNum);
             
              tft.setCursor(BATTERY_ICON_HORIZ , batteryNum);   // indent to horiz pixel location for battery icon     
              width1 =  3 + (i % (sizeof(batteryEmptyIcon)-3)) ; //start at offset of 3 pixels
              battery[batteryNum].oldVolt=battVoltage;
            }
           // Calculate the time duration between the last reading (in milliseconds)
            currentTime = millis();
            duration = (currentTime - battery[batteryNum].PrevTime);
            battery[batteryNum].PrevTime = currentTime;
            battery[batteryNum].RunTime += duration;
            // Current through resistor is voltage across the resistor divided by load restistance
            // Since the voltage is in millivolts, the current will be in milliamps
            float diff = battVoltage-fetVoltage;
            if(diff < 0) diff=0.0;
            loadCurrent = (battVoltage - fetVoltage) / LOAD_RESISTANCE;
            if(Serial){
            Serial.print("Vin: ");
            Serial.print(battVoltage);
            Serial.print("     Vout: ");
            Serial.print(fetVoltage);
            Serial.print("     Diff: ");
            Serial.print(diff);
            Serial.print("     Curr: ");
            Serial.println(loadCurrent/1000.0);
            }
            // milliAmpHours = current (in milliAmps)  * duration (in Hours)
            // Must divide by (60*60*1000) to convert duration in micro seconds to hours
            // But doing this now would cause a loss of precision, so
            // divide by 3600 which will result in microamp hours to be summed.
            // Divide by 1000 when milliamp hours are desired for display
            battery[batteryNum].charge +=   (loadCurrent * duration) / 3600;
           if(Serial ){
            Serial.print("Bat");
            Serial.print(batteryNum+1);
            Serial.print("  V=");
            Serial.print(battVoltage);
            Serial.print("   duration ");
            Serial.print(duration);
            Serial.print("ms   load=");
            Serial.print(loadCurrent );
            Serial.print("mA     mAh=");
            Serial.println(battery[batteryNum].charge/1000);
           }
           showGraph(batteryNum, battVoltage);
           
           if(battery[batteryNum].charge != battery[batteryNum].oldCap || battery[batteryNum].charge == 0.0){
              printCharge(battery[batteryNum].charge/1000,4,batteryNum);
              battery[batteryNum].oldCap=battery[batteryNum].charge;
           }
            // Has the battery voltage dropped below the minimum?
            // Must have several battery voltage samples below minimum
            // in a row, before declaring 'done'
            if (battVoltage <  battery[batteryNum].lowerThreshold)
            {
               if(Serial){
               Serial.print("Batt");
               Serial.print(batteryNum);
               Serial.print(" Below threshold: ");
               Serial.println(battVoltage);
               }
               if ( battery[batteryNum].numSamplesBelowMin < 3 ){
                  battery[batteryNum].numSamplesBelowMin++;
               } else {
                  // This testing on this battery is complete, set status to DONE
                  // Turn off the discharge load and update the display
                  battery[batteryNum].battStatus = DONE;
                  digitalWrite(battery[batteryNum].dischargeControlPin, LOW); // turn off the load
                  ClearDisplayLine(batteryNum);
                  printCharge(battery[batteryNum].charge/1000,4,batteryNum); 
                  printTime(batteryNum);
                  tone(SPKR_PIN, DONE_BEEP,50); // short beep
                  // Check to see if ALL installed batteries are in the DONE state.
                  done = checkAllDone();
               }
            } else {
              // reset counter since this reading was good.
             battery[batteryNum].numSamplesBelowMin = 0;
            } 
         } else if (battery[batteryNum].battStatus == DONE) {
            // don't update this line of the display    
           
         } else if (battery[batteryNum].battStatus == NOT_INSTALLED) {
             ClearDisplayLine(batteryNum);
             empty(batteryNum);
             if (battVoltage >= MIN_VOLTAGE){
                // This condition indicates that a battery has 
                // been installed, change status to Detecting
                battery[batteryNum].battStatus = DETECTING_TYPE;
             }   
         } else if (battery[batteryNum].battStatus == OVER_VOLTAGE ) {
            battery[batteryNum].battStatus =  DETECTING_TYPE;
         } else if (battery[batteryNum].battStatus == DETECTING_TYPE) {
            detectBatteryType(batteryNum, battVoltage );
         } else { // undefined battery status 
            // should never get here
         }  
      }
   } else {
      // we're done - all batteries are discharged
      if (beepCounter  < 3 )
      {
         tone(SPKR_PIN, DONE_BEEP,200);  // done beep
         beepCounter++;
      }
      tft.fillRect(215,50,106,37,YELLOW);
      tft.setTextColor(BLACK);
      tft.setFont(&FreeSans9pt7b);
      tft.setCursor(220,73);
      tft.print(" Time   mAh");
      // continue to update battery voltage
      for (batteryNum= 0 ; batteryNum < MAX_BATTERIES ; batteryNum++)
      {
        
         if (battery[batteryNum].battStatus == DONE)
         {
            // set cursor at beginning of line but don't erase battery icon
            // or the result mAh
            printTime(batteryNum);
         }
         else
         {
           battVoltage = getBatteryVoltage(batteryNum);
           ClearDisplayLine(batteryNum);
         }
      }
      
   }
    delay(1000);  // wait one second, then get next set of samples
} //end of main loop


void splash(){
  tft.fillScreen(BG_color);
  tft.drawRoundRect(0, 0, tft.width(), tft.height(),10, WHITE);
  tft.drawRoundRect(2, 2, tft.width()-4, tft.height()-4,10, WHITE);
  tft.setCursor(110, 30);
  
  tft.setTextColor(CYAN);
  tft.setTextSize(3);
  tft.println("Li-Ion");
  tft.setCursor(110, 85);
  tft.println("Battery");
  tft.setCursor(100, 120);
  tft.println("Capacity");
  tft.setCursor(120, 166);
  tft.println("Tester");
  tft.setTextSize(1);
  tft.setCursor(190, 220);
  tft.println("by mickey_zg");
  yield();
}

void mainScreen(){
  tft.drawRect(215,1,103,238,RED);
  tft.drawLine(215,50,318,50,YELLOW);
  tft.fillRect(215,50,318,57,YELLOW);
  tft.fillRect(215,88,318,238,BLACK);
  tft.setTextColor(BLACK);
  tft.setFont(&FreeSans9pt7b);
  tft.setCursor(220,73);
  tft.print("    V     mAh");
  tft.drawLine(268,88,268,238,WHITE);
  
  for (int batteryNum= 0 ; batteryNum < MAX_BATTERIES ; batteryNum++)
  {
    tft.setFont(&FreeSans9pt7b);
    tft.setCursor((batteryNum*50+15),20);
    tft.setTextColor(YELLOW);
    tft.print(batteryNum+1);
    tft.print(".");
  
    tft.setCursor(197,114+(batteryNum*38));
    tft.print(batteryNum+1);
    tft.print(".");
  }
    yield();
}

void printTime(int bat){
  
  unsigned long Now = battery[bat].RunTime/1000;
  int Seconds = Now%60;
  int Minutes = (Now/60)%60;
  int Hours = (Now/3600)%24;
  tft.setTextColor(BLACK);
  tft.setCursor(219,114+(bat*38));
  tft.print(String(Hours,DEC) + ":" + String(Minutes,DEC));

}
//////////////////////////////////////////////////////////////////////////////////
// Read analog input for specified battery (in millivolts)
//////////////////////////////////////////////////////////////////////////////////
unsigned int getBatteryVoltage(unsigned int batteryNum)
{
   float samples=0.0;
   float Voltage;
   for(int i=0; i<100;i++){
       samples=samples+analogRead(battery[batteryNum].batteryVoltagePin);
       delay(3);
   }
    return ((float)samples/100.0 * (float)refV / 1024.0)*battery[batteryNum].BatResRatio;
}

/////////////////////////////////////////////////////////////////////////////////
// Read analog input for specified battery's FET (in millivolts)
//////////////////////////////////////////////////////////////////////////////////
unsigned int getFetVoltage(unsigned int batteryNum)
{
   float samples=0.0;
   for(int i=0; i<100;i++){
      samples=samples+analogRead(battery[batteryNum].fetVoltagePin);
      delay(3);
   }
   return ((float)samples/100.0 * (float)refV / 1024.0)*battery[batteryNum].FetResRatio;
}

//////////////////////////////////////////////////////////////////////////////////
// printCharge() prints unsigned integer, right justified
//            on the LCD with the specified number of digits (up to 5)
//            supressing leading zeros. Prints asterisks if the number is too
//            big to be displayed.
//////////////////////////////////////////////////////////////////////////////////
void printCharge(unsigned int n, unsigned short numDigits, int bat)
{
  
  const unsigned int powersOfTen[]={1,10,100,1000,10000};
  boolean overflow = false, supressZero = true;
  unsigned int digit, d;
  if(battery[bat].battStatus == DONE){
    tft.setTextColor(BLACK);
    tft.fillRect(269,88+(bat*38),52,37,GREEN);
  } else {
    tft.setTextColor(WHITE);
    tft.fillRect(269,88+(bat*38),52,37,BLACK);
  }
  tft.setCursor(272,114+(bat*38));
  for (d = numDigits ; d > 0 ; d--)
  {
     if (overflow || numDigits > 5)
     {
      tft.print("*");
     }
     else
     {
       // pow() function doesn't work as expected - use array powersOfTen[]
       digit = n / powersOfTen[d-1];
       n = n % (powersOfTen[d-1]);
       if (digit == 0 && supressZero && d > 1){
          tft.print(" ");
       }else if (digit <= 9)
       {
          tft.print(digit);
          supressZero = false;
       }
       else
       {
         overflow = true;
         tft.print(" *** ");
       }
     }
  }          
}

//////////////////////////////////////////////////////////////////////////////////
// printVoltage() prints unsigned integer (millivolts), as a voltage with
//        decimal point on the LCD from 0.000 to 9.999 volts
//        Prints asterisks if the number is too big to be displayed.
//////////////////////////////////////////////////////////////////////////////////
void printVoltage(unsigned int n, int bat)
{
  
  const unsigned int powersOfTen[]={1,10,100,1000,10000},  numDigits = 4;
  boolean overflow = false, supressZero = true;
  unsigned int digit, d;
   if(battery[bat].battStatus == DONE){
    tft.setTextColor(BLACK);
    tft.fillRect(215,88+(bat*38),52,37,GREEN);
  } else {
    tft.setTextColor(WHITE);
    tft.fillRect(215,88+(bat*38),52,37,BLACK);
  }
  tft.setCursor(219,114+(bat*38));
  for (d = numDigits ; d > 0 ; d--)
  {
     if (overflow)
     {
      tft.print(" ***");
     }
     else
     {
       // pow() function doesn't work as expected - use array powersOfTen[]
       digit = n / powersOfTen[d-1];
       n = n % (powersOfTen[d-1]);
       if (digit <= 9)
       {
        tft.print(digit);
          if (d == numDigits){
             tft.print(".");
          }
       }
       else
       {
         overflow = true;
      tft.print(" ***");
       }
     }
  }          
}


//////////////////////////////////////////////////////////////////////////////////
// checkAllDone()  checks to see if ALL installed batteries are in the DONE state.
//                 return true if all tests are complete.
//////////////////////////////////////////////////////////////////////////////////
boolean checkAllDone()
{
  unsigned int batteryNum;
  unsigned int count=0;
  
   for (batteryNum= 0 ; batteryNum < MAX_BATTERIES ; batteryNum++)
   {
     if( battery[batteryNum].battStatus == TEST_IN_PROGRESS)
        return false;
   }
   return true;
}     

//////////////////////////////////////////////////////////////////////////////////
// detectBatteryType()  Detects the battery type and sets up the appropriate
//               status and thresholds
//////////////////////////////////////////////////////////////////////////////////
void detectBatteryType(unsigned int batteryNum, unsigned int battVoltage)
{
   if (battery[batteryNum].numSamplesAboveMin > 3)
   {
      if (battVoltage > MAX_VOLTAGE )
      {
         battery[batteryNum].battStatus = OVER_VOLTAGE;
      }
      else if (battVoltage > MIN_VOLTAGE)
      {
         // Battery Type identified as: li-ion
         // Initialize variables and start discharge
         battery[batteryNum].lowerThreshold = MIN_VOLTAGE;
         battery[batteryNum].battStatus = TEST_IN_PROGRESS;
         battery[batteryNum].charge = 0;
         battery[batteryNum].RunTime = 0;
         battery[batteryNum].PrevTime = millis();
         barMeter(batteryNum,schema);
         digitalWrite(battery[batteryNum].dischargeControlPin, HIGH); // turn the FET on
      }
      else
      {
         battery[batteryNum].numSamplesAboveMin = 0;
      }             
   }
   else // not enough good samples yet
   {
      if (battVoltage > MIN_VOLTAGE)
      {
         battery[batteryNum].numSamplesAboveMin++;
         battery[batteryNum].numSamplesBelowMin = 0;
      }
      else
      {
         battery[batteryNum].numSamplesBelowMin++;
         battery[batteryNum].numSamplesAboveMin = 0;
      }
   }
   if (battery[batteryNum].numSamplesBelowMin > 3)
   {
      battery[batteryNum].battStatus = NOT_INSTALLED;
   }      
}

//////////////////////////////////////////////////////////////////////////////////
// ClearDisplayLine()  utility function to clear bat line of the display
//////////////////////////////////////////////////////////////////////////////////
void ClearDisplayLine(int bat) 
{
  unsigned int i;
  tft.setCursor(300, 50+(bat*38));  // put cursor on first char of specified line
  if(battery[bat].battStatus == DONE){
    tft.fillRect(215,88+(bat*38),52,37,GREEN);
    tft.setTextColor(BLACK);
  } else{
    tft.fillRect(215,88+(bat*38),52,37,BLACK);
    tft.setTextColor(WHITE);
  }
  tft.setFont(&FreeSans9pt7b);
  tft.setTextSize(1);
}

void empty(int bat){
  tft.fillRect((bat*50+5), 35, 35, 200, CYAN);
}

void showGraph(int bat, float Volt){
  int  bHeight=map(Volt, MIN_VOLTAGE,MAX_VOLTAGE,36,200);
  tft.fillRect((bat*50+6),35,32,200-(bHeight),BG_color);
}

// #########################################################################
//  Draw the meter on the screen, returns x coord of righthand side
//  Function taken and adapted from https://www.instructables.com/id/Arduino-analogue-ring-meter-on-colour-TFT-display/
// #########################################################################
int barMeter(int bat, byte scheme)
{
  int text_colour = 0; // To hold the text colour
  byte inc = 4; // Draw segments every X px

  // Draw colour blocks every inc degrees
  for (int i = 200; i > 0; i -= inc) {

    // Choose colour from scheme
    int colour = 0;
    switch (scheme) {
      case 0: colour = ILI9341_RED; break; // Fixed colour
      case 1: colour = ILI9341_GREEN; break; // Fixed colour
      case 2: colour = ILI9341_BLUE; break; // Fixed colour
      case 3: colour = rainbow(map(i, 0, 200, 0, 127)); break; // Full spectrum blue to red*/
      case 4: colour = rainbow(map(i, 0, 200, 73, 110)); break; // Green to red (high temperature etc)
      
      case 5: colour = rainbow(map(i, 0, 200, 110, 73)); break; // Red to green (low battery etc)
      default: colour = ILI9341_BLUE; break; // Fixed colour
    }

    uint16_t x0 = (bat*50+5); //sx * (r - w) + x;
    uint16_t y0 = 34 ;//sy * (r - w) + y;
    uint16_t x1 = 34;//sx * r + x;
    uint16_t y1 = i;
    if (i < 201) { // Fill in coloured segments
      tft.fillRect(x0, y0, x1, y1, colour);
      text_colour = colour; // Save the last colour drawn
    } else { // Fill in blank segments
      tft.fillRect(x0, y0, x1, y1, BG_color);
    }
  }
}
// #########################################################################
// Return a 16 bit rainbow colour
// #########################################################################
unsigned int rainbow(byte value)
{
  // Value is expected to be in range 0-127
  // The value is converted to a spectrum colour from 0 = blue through to 127 = red

  byte red = 0; // Red is the top 5 bits of a 16 bit colour value
  byte green = 0;// Green is the middle 6 bits
  byte blue = 0; // Blue is the bottom 5 bits

  byte quadrant = value / 32;

  if (quadrant == 0) {
    blue = 31;
    green = 2 * (value % 32);
    red = 0;
  }
  if (quadrant == 1) {
    blue = 31 - (value % 32);
    green = 63;
    red = 0;
  }
  if (quadrant == 2) {
    blue = 0;
    green = 63;
    red = value % 32;
  }
  if (quadrant == 3) {
    blue = 0;
    green = 63 - 2 * (value % 32);
    red = 31;
  }
  return (red << 11) + (green << 5) + blue;
}

long readVcc() {
  long result;
  // Read 1.1V reference against AVcc
#if defined(__AVR_ATmega32U4__) || defined(__AVR_ATmega1280__) || defined(__AVR_ATmega2560__)
  ADMUX = _BV(REFS0) | _BV(MUX4) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
#elif defined (__AVR_ATtiny24__) || defined(__AVR_ATtiny44__) || defined(__AVR_ATtiny84__)
  ADMUX = _BV(MUX5) | _BV(MUX0);
#elif defined (__AVR_ATtiny25__) || defined(__AVR_ATtiny45__) || defined(__AVR_ATtiny85__)
  ADMUX = _BV(MUX3) | _BV(MUX2);
#else
  ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
#endif
  delay(2); // Wait for Vref to settle
  ADCSRA |= _BV(ADSC); // Convert
  while (bit_is_set(ADCSRA, ADSC));
  result = ADCL;
  result |= ADCH << 8;
  result = 1126400L / result; // Calculate Vcc (in mV); 1126400 = 1.1*1024*1000
  return result;
}
